;;; This software is being provided to you, the LICENSEE, by the
;;; Massachusetts Institute of Technology (M.I.T.), and the NEC Research
;;; Institute (NECI) under the following license.  By obtaining, using
;;; and/or copying this software, you agree that you have read,
;;; understood, and will comply with these terms and conditions:
;;; 
;;; Permission to use, copy, modify and distribute, including the right to
;;; grant others rights to distribute at any tier, this software and its
;;; documentation for research purposes only and without fee or royalty is
;;; hereby granted, provided that you agree to comply with the following
;;; copyright notice and statements, including the disclaimer, and that
;;; the same appear on ALL copies of the software and documentation,
;;; including modifications that you make for internal use or for
;;; distribution:
;;; 
;;; Copyright 1995 by the Massachusetts Institute of Technology.  All
;;; rights reserved.  Copyright 1995 by NEC Research Institute Inc. All
;;; rights reserved
;;; 
;;; THIS SOFTWARE IS PROVIDED "AS IS", AND M.I.T. and NECI MAKE NO
;;; REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  By way of example,
;;; but not limitation, M.I.T. and NECI MAKE NO REPRESENTATIONS OR
;;; WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR
;;; THAT THE USE OF THE LICENSED SOFTWARE OR DOCUMENTATION WILL NOT
;;; INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER
;;; RIGHTS.
;;; 
;;; The name of the Massachusetts Institute of Technology or M.I.T., or
;;; NEC Research Institute, or NECI may NOT be used in advertising or
;;; publicity pertaining to distribution of the software.  Title to
;;; copyright in this software and any associated documentation shall at
;;; all times remain with M.I.T. and NECI, and USER agrees to preserve
;;; same.
;;; 

;;; -*- Mode: LISP; Base: 10; Package: G; Syntax: Common-Lisp -*-
(in-package G)

(export '(*short-pi* *2pi* *short-2pi* *pi/2* *short-pi/2* *pi/4* *short-pi/4* *epsilon* *-epsilon*))

(defconstant *SHORT-PI* (coerce pi 'short-float) "PI as a short-float.")

(defconstant *2PI* (* 2.0 pi) "2*PI.")
(defconstant *SHORT-2PI* (coerce (* 2.0 pi) 'short-float) "2*PI as a short-float.")

(defconstant *PI/2* (/ pi 2.0) "PI / 2.")
(defconstant *SHORT-PI/2* (coerce (/ pi 2.0) 'short-float) "PI / 2 as a short-float.")

(defconstant *PI/4* (/ pi 4.0) "PI / 4.")
(defconstant *SHORT-PI/4* (coerce (/ pi 4.0) 'short-float) "PI / 4 as a short-float.")

(defvar *epsilon* 0.0001 "Tolerance used in almost-equal calculations. (Should always be positive.)")
(defvar *-epsilon* (- *epsilon*))

(defconstant *SQRT2* (sqrt 2.0) "The square root of 2")

